@@ -10,16 +10,26 @@ import {
   incrRoundUp,
   TimeRange,
   FieldType,
+  getDisplayProcessor,
 } from '@grafana/data';
 import { AxisPlacement, ScaleDirection, ScaleDistribution, ScaleOrientation, HeatmapCellLayout } from '@grafana/schema';
-import { UPlotConfigBuilder } from '@grafana/ui';
-import { isHeatmapCellsDense, readHeatmapRowsCustomMeta } from 'app/features/transformers/calculateHeatmap/heatmap';
+import { UPlotConfigBuilder, UPlotConfigPrepFn } from '@grafana/ui';
+import {
+  calculateBucketFactor,
+  isHeatmapCellsDense,
+  readHeatmapRowsCustomMeta,
+} from 'app/features/transformers/calculateHeatmap/heatmap';
 
 import { pointWithin, Quadtree, Rect } from '../barchart/quadtree';
 
 import { HeatmapData } from './fields';
 import { FieldConfig, HeatmapSelectionMode, YAxisConfig } from './types';
 
+/** Validates and returns a safe log base (2 or 10), defaults to 2 if invalid */
+export function toLogBase(value: number | undefined): 2 | 10 {
+  return value === 10 ? 10 : 2;
+}
+
 interface PathbuilderOpts {
   each: (u: uPlot, seriesIdx: number, dataIdx: number, lft: number, top: number, wid: number, hgt: number) => void;
   gap?: number | null;
@@ -52,6 +62,8 @@ interface PrepConfigOpts {
   yAxisConfig: YAxisConfig;
   ySizeDivisor?: number;
   selectionMode?: HeatmapSelectionMode;
+  xAxisConfig?: Parameters<UPlotConfigPrepFn>[0]['xAxisConfig'];
+  rowsFrame?: { yBucketScale?: { type: ScaleDistribution; log?: number; linearThreshold?: number } };
 }
 
 export function prepConfig(opts: PrepConfigOpts) {
@@ -66,8 +78,12 @@ export function prepConfig(opts: PrepConfigOpts) {
     yAxisConfig,
     ySizeDivisor,
     selectionMode = HeatmapSelectionMode.X,
+    xAxisConfig,
+    rowsFrame,
   } = opts;
 
+  const yBucketScale = rowsFrame?.yBucketScale;
+
   const xScaleKey = 'x';
   let isTime = true;
 
@@ -163,6 +179,13 @@ export function prepConfig(opts: PrepConfigOpts) {
     }
   }
 
+  let xField = dataRef.current?.heatmap?.fields[0]!;
+  xField.display ??= getDisplayProcessor({
+    field: xField,
+    theme,
+    timeZone,
+  });
+
   builder.addAxis({
     scaleKey: xScaleKey,
     placement: AxisPlacement.Bottom,
@@ -170,6 +193,11 @@ export function prepConfig(opts: PrepConfigOpts) {
     isTime,
     theme: theme,
     timeZone,
+    formatValue:
+      isTime && xField.config.unit?.startsWith('time:')
+        ? (v, decimals) => xField.display!(v, decimals).text
+        : undefined,
+    ...xAxisConfig,
   });
 
   const yField = dataRef.current?.heatmap?.fields[1]!;
@@ -181,7 +209,20 @@ export function prepConfig(opts: PrepConfigOpts) {
   const yScale = yFieldConfig?.scaleDistribution ?? { type: ScaleDistribution.Linear };
   const yAxisReverse = Boolean(yAxisConfig.reverse);
   const isSparseHeatmap = heatmapType === DataFrameType.HeatmapCells && !isHeatmapCellsDense(dataRef.current?.heatmap!);
-  const shouldUseLogScale = yScale.type !== ScaleDistribution.Linear || isSparseHeatmap;
+
+  const scaleDistribution = (() => {
+    if (yBucketScale) {
+      return yBucketScale.type;
+    }
+    if (yScale.type !== ScaleDistribution.Linear || isSparseHeatmap) {
+      return ScaleDistribution.Log;
+    }
+    return ScaleDistribution.Linear;
+  })();
+
+  const scaleLog = toLogBase(yBucketScale?.log ?? yScale.log);
+  const scaleLinearThreshold = yBucketScale?.linearThreshold;
+
   const isOrdinalY = readHeatmapRowsCustomMeta(dataRef.current?.heatmap).yOrdinalDisplay != null;
 
   // random to prevent syncing y in other heatmaps
@@ -195,30 +236,36 @@ export function prepConfig(opts: PrepConfigOpts) {
     orientation: ScaleOrientation.Vertical,
     direction: yAxisReverse ? ScaleDirection.Down : ScaleDirection.Up,
     // should be tweakable manually
-    distribution: shouldUseLogScale ? ScaleDistribution.Log : ScaleDistribution.Linear,
-    log: yScale.log ?? 2,
+    distribution: scaleDistribution,
+    log: scaleLog,
+    linearThreshold: scaleLinearThreshold,
     range:
       // sparse already accounts for le/ge by explicit yMin & yMax cell bounds, so no need to expand y range
       isSparseHeatmap
         ? (u, dataMin, dataMax) => {
+            // Extract yMin and yMax arrays
+            const yMinData = u.data[1]?.[1];
+            const yMaxData = u.data[1]?.[2];
+            const yMinValues = Array.isArray(yMinData) ? yMinData : [];
+            const yMaxValues = Array.isArray(yMaxData) ? yMaxData : [];
+
             // ...but uPlot currently only auto-ranges from the yMin facet data, so we have to grow by 1 extra factor
-            // @ts-ignore
-            let bucketFactor = u.data[1][2][0] / u.data[1][1][0];
+            const bucketFactor = calculateBucketExpansionFactor(yMinValues, yMaxValues);
 
             dataMax *= bucketFactor;
 
             let scaleMin: number | null, scaleMax: number | null;
 
-            [scaleMin, scaleMax] = shouldUseLogScale
-              ? uPlot.rangeLog(dataMin, dataMax, (yScale.log ?? 2) as unknown as uPlot.Scale.LogBase, true)
-              : [dataMin, dataMax];
+            const isLogScale =
+              scaleDistribution === ScaleDistribution.Log || scaleDistribution === ScaleDistribution.Symlog;
+            [scaleMin, scaleMax] = isLogScale ? uPlot.rangeLog(dataMin, dataMax, scaleLog, true) : [dataMin, dataMax];
 
-            if (shouldUseLogScale && !isOrdinalY) {
+            let { min: explicitMin, max: explicitMax } = yAxisConfig;
+
+            if (isLogScale && !isOrdinalY) {
               let yExp = u.scales[yScaleKey].log!;
               let log = yExp === 2 ? Math.log2 : Math.log10;
 
-              let { min: explicitMin, max: explicitMax } = yAxisConfig;
-
               // guard against <= 0
               if (explicitMin != null && explicitMin > 0) {
                 // snap to magnitude
@@ -230,6 +277,9 @@ export function prepConfig(opts: PrepConfigOpts) {
                 let maxLog = log(explicitMax);
                 scaleMax = yExp ** incrRoundUp(maxLog, 1);
               }
+            } else if (!isOrdinalY) {
+              // Apply explicit min/max for linear scale
+              [scaleMin, scaleMax] = applyExplicitMinMax(scaleMin, scaleMax, explicitMin, explicitMax);
             }
 
             return [scaleMin, scaleMax];
@@ -242,7 +292,7 @@ export function prepConfig(opts: PrepConfigOpts) {
             let { min: explicitMin, max: explicitMax } = yAxisConfig;
 
             // logarithmic expansion
-            if (shouldUseLogScale) {
+            if (scaleDistribution === ScaleDistribution.Log || scaleDistribution === ScaleDistribution.Symlog) {
               let yExp = u.scales[yScaleKey].log!;
 
               let minExpanded = false;
@@ -265,17 +315,31 @@ export function prepConfig(opts: PrepConfigOpts) {
                 }
               }
 
+              // For pre-bucketed data with explicit scale, calculate expansion factor from actual bucket spacing
+              // For calculated heatmaps, use the full log base
+              let expansionFactor: number = yExp;
+
+              if (yBucketScale !== undefined) {
+                // Try to infer the bucket factor from the actual data spacing
+                const yValues = u.data[1]?.[1];
+                if (Array.isArray(yValues) && yValues.length >= 2 && typeof yValues[0] === 'number') {
+                  expansionFactor = calculateBucketFactor(yValues, yExp);
+                }
+              }
+
               if (dataRef.current?.yLayout === HeatmapCellLayout.le) {
                 if (!minExpanded) {
-                  scaleMin /= yExp;
+                  scaleMin /= expansionFactor;
                 }
               } else if (dataRef.current?.yLayout === HeatmapCellLayout.ge) {
                 if (!maxExpanded) {
-                  scaleMax *= yExp;
+                  scaleMax *= expansionFactor;
                 }
               } else {
-                scaleMin /= yExp / 2;
-                scaleMax *= yExp / 2;
+                // Unknown layout - expand both directions
+                const factor = Math.sqrt(expansionFactor); // Use sqrt for balanced expansion
+                scaleMin /= factor;
+                scaleMax *= factor;
               }
 
               if (!isOrdinalY) {
@@ -368,7 +432,7 @@ export function prepConfig(opts: PrepConfigOpts) {
             return splits.map((v) =>
               v < 0
                 ? (meta.yMinDisplay ?? '') // Check prometheus style labels
-                : (meta.yOrdinalDisplay[v] ?? '')
+                : (meta.yOrdinalDisplay?.[v] ?? '')
             );
           }
           return splits;
@@ -411,11 +475,15 @@ export function prepConfig(opts: PrepConfigOpts) {
           : dataRef.current?.xLayout === HeatmapCellLayout.ge
             ? 1
             : 0,
-      yAlign: ((dataRef.current?.yLayout === HeatmapCellLayout.le
-        ? -1
-        : dataRef.current?.yLayout === HeatmapCellLayout.ge
-          ? 1
-          : 0) * (yAxisReverse ? -1 : 1)) as -1 | 0 | 1,
+      yAlign: (() => {
+        const yAlign =
+          dataRef.current?.yLayout === HeatmapCellLayout.le
+            ? -1
+            : dataRef.current?.yLayout === HeatmapCellLayout.ge
+              ? 1
+              : 0;
+        return yAxisReverse ? (yAlign === -1 ? 1 : yAlign === 1 ? -1 : 0) : yAlign;
+      })(),
       ySizeDivisor,
       disp: {
         fill: {
@@ -566,15 +634,19 @@ export function heatmapPathsDense(opts: PathbuilderOpts) {
         let ySize: number;
 
         if (scaleX.distr === 3) {
-          xSize = Math.abs(valToPosX(xs[0] * scaleX.log!, scaleX, xDim, xOff) - valToPosX(xs[0], scaleX, xDim, xOff));
+          // For log scales, calculate cell size from actual adjacent bucket positions
+          const nextXValue = xs[yBinQty] ?? xs[0] * scaleX.log!;
+          xSize = Math.abs(valToPosX(nextXValue, scaleX, xDim, xOff) - valToPosX(xs[0], scaleX, xDim, xOff));
         } else {
           xSize = Math.abs(valToPosX(xBinIncr, scaleX, xDim, xOff) - valToPosX(0, scaleX, xDim, xOff));
         }
 
         if (scaleY.distr === 3) {
-          ySize =
-            Math.abs(valToPosY(ys[0] * scaleY.log!, scaleY, yDim, yOff) - valToPosY(ys[0], scaleY, yDim, yOff)) /
-            ySizeDivisor;
+          // Use actual data spacing for pre-bucketed data, or full magnitude for calculated heatmaps with splits
+          const nextYValue = ySizeDivisor === 1 ? (ys[1] ?? ys[0] * scaleY.log!) : ys[0] * scaleY.log!;
+
+          const baseYSize = Math.abs(valToPosY(nextYValue, scaleY, yDim, yOff) - valToPosY(ys[0], scaleY, yDim, yOff));
+          ySize = baseYSize / ySizeDivisor;
         } else {
           ySize = Math.abs(valToPosY(yBinIncr, scaleY, yDim, yOff) - valToPosY(0, scaleY, yDim, yOff)) / ySizeDivisor;
         }
@@ -815,6 +887,32 @@ export function heatmapPathsSparse(opts: PathbuilderOpts) {
   };
 }
 
+/**
+ * Calculates a bucket expansion factor from yMin/yMax data arrays.
+ * Used to expand the y-axis range for sparse heatmaps where uPlot only auto-ranges from yMin.
+ *
+ * @param yMinValues - Array of yMin bucket boundary values
+ * @param yMaxValues - Array of yMax bucket boundary values
+ * @returns A valid expansion factor, or 1 as fallback
+ */
+export function calculateBucketExpansionFactor(yMinValues: unknown[], yMaxValues: unknown[]): number {
+  // Guard against invalid bucket factors (e.g., division by zero when first bucket starts at 0)
+  for (let i = 0; i < yMinValues.length; i++) {
+    const yMin = yMinValues[i];
+    const yMax = yMaxValues[i];
+    if (typeof yMin !== 'number' || typeof yMax !== 'number') {
+      continue;
+    }
+    const factor = yMax / yMin;
+    // finite checks for divide-by-zero
+    if (Number.isFinite(factor) && factor > 0) {
+      return factor;
+    }
+  }
+
+  return 1;
+}
+
 export const boundedMinMax = (
   values: number[],
   minValue?: number,
@@ -863,3 +961,30 @@ export const valuesToFills = (values: number[], palette: string[], minValue: num
 
   return indexedFills;
 };
+
+/**
+ * Calculates the Y-axis size divisor for heatmap cell rendering.
+ * For log/symlog scales with calculated data (no explicit scale), divides cells by the split value.
+ * Otherwise returns 1 (no division).
+ */
+export function calculateYSizeDivisor(
+  scaleType: ScaleDistribution | undefined,
+  hasExplicitScale: boolean,
+  splitValue: number | string | undefined
+): number {
+  const isLogScale = scaleType === ScaleDistribution.Log || scaleType === ScaleDistribution.Symlog;
+  return isLogScale && !hasExplicitScale ? +(splitValue || 1) : 1;
+}
+
+/**
+ * Applies explicit min/max values to scale range for linear scales.
+ * Returns the original values if explicitMin/explicitMax are undefined.
+ */
+export function applyExplicitMinMax(
+  scaleMin: number | null,
+  scaleMax: number | null,
+  explicitMin: number | undefined,
+  explicitMax: number | undefined
+): [number | null, number | null] {
+  return [explicitMin ?? scaleMin, explicitMax ?? scaleMax];
+}
